/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on April 29, 2010, 7:31 PM
 */

#ifndef _U2D_PROJECT_H
#define	_U2D_PROJECT_H

#include <boost/filesystem.hpp>
#include <string>

#include "exceptions.h"

namespace bfs = boost::filesystem;

class u2d_Project {
public:
//    u2d_Project();
    u2d_Project(string project_name);
    virtual ~u2d_Project();

    void setProjectName(string project_name);
    std::string getProjectName() const;

    void setProjectRelativeDir(string project_relative_dir);
    std::string getProjectRelativeDir(bool withName = false) const;

    void setProjectAbsoluteDir(string project_absolute_dir);
    std::string getProjectAbsoluteDir(bool withName = false) const;


    std::string getMeshNodesListFilename(bool path = true, bool rel = false) const;

    std::string getMeshGridListFilename(bool path = true, bool rel = false) const;

    std::string getMeshDomainFilename(bool path = true, bool rel = false) const;

    std::string getMeshBoundaryFilename(bool path = true, bool rel = false) const;
    
    std::string getMeshGridplotFilename(bool path = true, bool rel = false) const;
    
    
    void updateRelativePath();

    void dump();
    
    static u2d_Project *CreateProject(bfs::path project_path_and_name)
    throw (u2d_IO_Exception &);
    static u2d_Project *CreateProject(bfs::path project_path, bfs::path project_name)
    throw (u2d_IO_Exception &);
    static u2d_Project *LoadProject(bfs::path project_path_and_name)
    throw (u2d_IO_Exception &);
    static u2d_Project *LoadProject(bfs::path project_path, bfs::path project_name)
    throw (u2d_IO_Exception &);

private:
    /* boost::filesystem path */
    //bfs::path path;
    /* The project name */
    std::string project_name;
    /* The absolute dir where the project files are located.
     * With a training slash */
    std::string project_absolute_dir;
    /* The dir relative to the execution point where the
     * project files are located
     * With a training slash */
    std::string project_relative_dir;

    /* A naive uncomplete function for boost path */
     static boost::filesystem::path NaiveUncomplete(boost::filesystem::path const path,
        boost::filesystem::path const base);
     static boost::filesystem::path RealPath(boost::filesystem::path const originalPath);
     
     
};

#endif	/* _U2D_PROJECT_H */

