/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2011 Gabriele Dini Ciacci

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Created on April 27, 2011, 5:37 PM
 */


#include "u2d_ValuesOutput.h"

#include "boost/numeric/mtl/mtl.hpp"

u2d_ValuesOutput::u2d_ValuesOutput() {
}

u2d_ValuesOutput::u2d_ValuesOutput(string _filename) :
filename(_filename) {
}

u2d_ValuesOutput::~u2d_ValuesOutput() {
}


void u2d_ValuesOutput::setFilename(string _filename)
    throw (u2d_Parse_Error_Exception &) {
    if(_filename.empty() == true) {
        throw u2d_Parse_Error_Exception("Filename not specified name is empty");
    }
    filename = _filename;
}

string u2d_ValuesOutput::getFilename() {
    return filename;
}

void u2d_ValuesOutput::openFile()
    throw (u2d_IO_Exception &) {
    file = fopen(filename.c_str(), "w+");
    if(file == NULL) {
        throw u2d_IO_Exception("Can't open vector file for output\n");
    }
}

void u2d_ValuesOutput::closeFile()
    throw (u2d_IO_Exception &) {
    int ret = 0;
    ret = fclose(file);
    if(ret == EOF) {
        throw u2d_IO_Exception("Can't close vector output file\n");
    }
}

/**
 * Prints a double value per line with exponential notation.
 * @param scalar The first double
 * @remark This should not be declared inline because it is used by inherited
 * (and is placed in a .cpp instead of a .h).
 */
void u2d_ValuesOutput::printScalarValueLine(double scalar) {
    fprintf(file, "%e\n", scalar);
}

/**
 * Prints a two double values per line with exponential notation.
 * @param x The first double
 * @param y The second double
 * @remark This should not be declared inline because it is used by inherited
 * (and is placed in a .cpp instead of a .h).
 */
void u2d_ValuesOutput::print2DVectorValueLine(double x, double y) {
    fprintf(file, "%e %e\n", x, y);
}

/**
 * Prints a threee double values per line with exponential notation.
 * @param x The first double
 * @param y The second double
 * @param z The third double
 * @remark This should not be declared inline because it is used by inherited
 * (and is placed in a .cpp instead of a .h).
 */
void u2d_ValuesOutput::print3DVectorValueLine(double x, double y, double z) {
    fprintf(file, "%e %e %e\n", x, y, z);
}