MODULE grid_types

   IMPLICIT NONE

   INTEGER, PARAMETER :: maxp = 2, maxs = maxp+1

   INTEGER, PARAMETER :: nul_cnd = 0, nul_par = 0
   INTEGER, PARAMETER :: frz_status = 0, und_status = 1,   &
                     &   int_status = 2, ext_status = 3,   &
                     &   stc_status = 4

   REAL(KIND=8), PARAMETER :: zero = 0.0d0
   REAL(KIND=8), PARAMETER :: one  = 1.0d0

   TYPE opposits
      TYPE (point),   POINTER :: pnt
      TYPE (simplex), POINTER :: spx
   END TYPE opposits

   TYPE point
      LOGICAL       :: visited
      INTEGER       :: index, cnd, ibk, parent
      REAL (KIND=8) :: leng

      TYPE (point),   POINTER :: prev, next, old, ppp
      TYPE (simplex), POINTER :: spx

      REAL (KIND=8), DIMENSION (maxp)      :: x
      REAL (KIND=8), DIMENSION (2*maxp-1)  :: ellipse
      REAL (KIND=8), DIMENSION (maxp,maxp) :: metric
   END TYPE point

   TYPE simplex
      LOGICAL       :: visited
      INTEGER       :: index, status
      REAL (KIND=8) :: det

      TYPE (simplex), POINTER :: prev, next, int, ext

      INTEGER,         DIMENSION (maxs,maxs) :: cmat
      REAL (KIND=8),   DIMENSION (maxs,maxs) :: fmat
      TYPE (opposits), DIMENSION (maxs)      :: opp
   END TYPE simplex
!######    CMAT      #####
!
!i simplessi adiacenti(opposti,3) di un simplesso si identificano tramite 
!un indice locale "i" da usare in opp(i)
!per raggiungere velocemente i 9 adiacenti dei 3 opposti serve
!la numerazione locale dei 3 opposti.
!cmat(i,j)
!colonna j ->informazioni sul simplesso opposto al nodo che localmente 
!            ha indice j
!cmat(j,j)=numero locale del nodo che sta di fronte a j
!cmat(i,j)=numero locale del nodo i nella numerazione locale del simplesso 
!          opposto a j
!          /\
!         / 3\       **numerazione antioraria
!        /    \ 
!       /1    2\   ->   cmat = .  .  .
!       --------               .  .  . 
!       \1    3/               1  3  2       
!        \    /            
!         \2 /
!          \/             

CONTAINS


   FUNCTION new_point () RESULT (p)

      IMPLICIT NONE

      TYPE (point), POINTER :: p

      INTEGER :: i


      ALLOCATE ( p )

      p % visited = .FALSE.
      p % index   = 0
      p % cnd     = nul_cnd
      p % ibk     = 0
      p % parent  = nul_par

      NULLIFY ( p % prev )
      NULLIFY ( p % next )
      NULLIFY ( p % old )
      NULLIFY ( p % ppp )
      NULLIFY ( p % spx )

      p % leng    = zero
      p % x       = zero
      p % ellipse = zero
      p % metric  = zero

      DO i=1,maxp
         p % metric(i,i)  = one
      ENDDO

   END FUNCTION new_point


   FUNCTION new_simplex () RESULT (s)

      IMPLICIT NONE

      TYPE (simplex), POINTER :: s

      INTEGER :: i


      ALLOCATE ( s )

      s % visited = .FALSE.
      s % index   = 0
      s % status  = und_status
      s % det     = zero

     NULLIFY ( s % prev )
      NULLIFY ( s % next )
      NULLIFY ( s % int )
      NULLIFY ( s % ext )

      DO i=1,maxs
	      NULLIFY (s % opp(i) % pnt)
	      NULLIFY (s % opp(i) % spx)
      ENDDO

      s % cmat    = 0
      s % fmat    = zero

   END FUNCTION new_simplex


END MODULE grid_types
